@@ -32,6 +32,8 @@
 
 #include <string.h>
 #include <stdio.h>
+#include <assert.h>
+#include <math.h>
 
 /* X and server generic header files */
 #include "xf86.h"
@@ -67,18 +69,12 @@ RADEONInitDispBandwidthAVIVO(ScrnInfoPtr pScrn,
 			     DisplayModePtr mode2, int pixel_bytes2);
 
 void
-radeon_crtc_dpms(xf86CrtcPtr crtc, int mode)
+radeon_do_crtc_dpms(xf86CrtcPtr crtc, int mode)
 {
     RADEONInfoPtr info = RADEONPTR(crtc->scrn);
     RADEONEntPtr pRADEONEnt = RADEONEntPriv(crtc->scrn);
-    RADEONCrtcPrivatePtr radeon_crtc = crtc->driver_private;
     xf86CrtcPtr crtc0 = pRADEONEnt->pCrtc[0];
-
-    if ((mode == DPMSModeOn) && radeon_crtc->enabled)
-	return;
-
-    if (mode == DPMSModeOff)
-	radeon_crtc_modeset_ioctl(crtc, FALSE);
+    RADEONCrtcPrivatePtr radeon_crtc = crtc->driver_private;
 
     if (IS_AVIVO_VARIANT || info->r4xx_atom) {
 	atombios_crtc_dpms(crtc, mode);
@@ -99,6 +95,20 @@ radeon_crtc_dpms(xf86CrtcPtr crtc, int mode)
 		legacy_crtc_dpms(crtc0, mode);
 	}
     }
+}
+
+void
+radeon_crtc_dpms(xf86CrtcPtr crtc, int mode)
+{
+    RADEONCrtcPrivatePtr radeon_crtc = crtc->driver_private;
+
+    if ((mode == DPMSModeOn) && radeon_crtc->enabled)
+	return;
+
+    if (mode == DPMSModeOff)
+	radeon_crtc_modeset_ioctl(crtc, FALSE);
+
+    radeon_do_crtc_dpms(crtc, mode);
 
     if (mode != DPMSModeOff) {
 	radeon_crtc_modeset_ioctl(crtc, TRUE);
@@ -237,7 +247,7 @@ RADEONComputePLL_old(RADEONPLLPtr pll,
 
 		    if (flags & RADEON_PLL_PREFER_CLOSEST_LOWER) {
 			error = freq - current_freq;
-			error = error < 0 ? 0xffffffff : error;
+			error = (int32_t)error < 0 ? 0xffffffff : error;
 		    } else
 			error = abs(current_freq - freq);
 		    vco_diff = abs(vco - best_vco);
@@ -508,6 +518,8 @@ radeon_crtc_load_lut(xf86CrtcPtr crtc)
     if (!crtc->enabled)
 	return;
 
+    radeon_save_palette_on_demand(pScrn, radeon_crtc->crtc_id);
+
     if (IS_DCE4_VARIANT) {
 	OUTREG(EVERGREEN_DC_LUT_CONTROL + radeon_crtc->crtc_offset, 0);
 
@@ -556,7 +568,6 @@ radeon_crtc_load_lut(xf86CrtcPtr crtc)
 	if (IS_AVIVO_VARIANT)
 	    OUTREG(AVIVO_D1GRPH_LUT_SEL + radeon_crtc->crtc_offset, radeon_crtc->crtc_id);
     }
-
 }
 
 static void
@@ -885,7 +896,7 @@ Bool RADEONAllocateControllers(ScrnInfoPtr pScrn, int mask)
     }
 
     /* 6 crtcs on DCE4 chips */
-    if (IS_DCE4_VARIANT && ((mask & 3) == 3)) {
+    if (IS_DCE4_VARIANT && ((mask & 3) == 3) && !IS_DCE41_VARIANT) {
 	for (i = 2; i < RADEON_MAX_CRTC; i++) {
 	    pRADEONEnt->pCrtc[i] = xf86CrtcCreate(pScrn, &radeon_crtc_funcs);
 	    if (!pRADEONEnt->pCrtc[i])
